@@ -61,6 +61,7 @@ const struct value_string osmo_mgcpc_codec_names[] = {
 	{ CODEC_AMRWB_16000_1, "AMR-WB/16000/1" },
 	{ CODEC_IUFP, "VND.3GPP.IUFP/16000" },
 	{ CODEC_CLEARMODE, "CLEARMODE/8000" },
+	{ CODEC_COMPR_CSD, "VND.THEMWI.CCSD/8000" },
 	{ 0, NULL },
 };
 
@@ -786,8 +787,10 @@ static void mgcp_read_cb(struct osmo_io_fd *iofd, int res, struct msgb *msg)
 	struct mgcp_client *mgcp = osmo_iofd_get_data(iofd);
 
 	if (res <= 0) {
+		char errbuf[128] = "";
+		strerror_r(-res, errbuf, sizeof(errbuf));
 		LOGPMGW(mgcp, LOGL_ERROR, "Failed to read: %s: %d='%s'\n",
-			osmo_iofd_get_name(iofd), res, strerror(res));
+			osmo_iofd_get_name(iofd), res, errbuf);
 
 		msgb_free(msg);
 		return;
@@ -822,8 +825,10 @@ static void mgcp_write_cb(struct osmo_io_fd *iofd, int res, struct msgb *msg)
 	struct mgcp_client *mgcp = osmo_iofd_get_data(iofd);
 
 	if (OSMO_UNLIKELY(res != msg->len)) {
+		char errbuf[128] = "";
+		strerror_r(-res, errbuf, sizeof(errbuf));
 		LOGPMGW(mgcp, LOGL_ERROR, "Failed to Tx MGCP: %s: %d='%s'; msg: len=%u '%s'...\n",
-			osmo_iofd_get_name(mgcp->iofd), res, strerror(res),
+			osmo_iofd_get_name(mgcp->iofd), res, errbuf,
 			msg->len, osmo_escape_str((const char *)msg->data, OSMO_MIN(42, msg->len)));
 	}
 }
@@ -860,6 +865,10 @@ static void _mgcp_client_send_dlcx(struct mgcp_client *mgcp, const char *epname)
 	};
 	osmo_strlcpy(mgcp_msg_dlcx.endpoint, epname, sizeof(mgcp_msg_dlcx.endpoint));
 	msgb_dlcx = mgcp_msg_gen(mgcp, &mgcp_msg_dlcx);
+	if (!msgb_dlcx) {
+		LOGPMGW(mgcp, LOGL_ERROR, "Failed generating MGCP DLCX %s\n", epname);
+		return;
+	}
 	mgcp_client_tx(mgcp, msgb_dlcx, &_ignore_mgcp_response, NULL);
 }
 
@@ -873,6 +882,10 @@ static void _mgcp_client_send_auep(struct mgcp_client *mgcp, const char *epname)
 	};
 	OSMO_STRLCPY_ARRAY(mgcp_msg_auep.endpoint, epname);
 	msgb_auep = mgcp_msg_gen(mgcp, &mgcp_msg_auep);
+	if (!msgb_auep) {
+		LOGPMGW(mgcp, LOGL_ERROR, "Failed generating MGCP AUEP %s\n", epname);
+		return;
+	}
 	mgcp_client_tx(mgcp, msgb_auep, &_ignore_mgcp_response, NULL);
 }
 
@@ -1047,6 +1060,7 @@ void mgcp_client_disconnect(struct mgcp_client *mgcp)
 	osmo_iofd_txqueue_clear(mgcp->iofd);
 	LOGPMGW(mgcp, LOGL_INFO, "MGCP association: %s -- closed!\n", osmo_iofd_get_name(mgcp->iofd));
 	osmo_iofd_free(mgcp->iofd);
+	mgcp->iofd = NULL;
 }
 
 /*! Get the IP-Aaddress of the associated MGW as string.
@@ -1307,6 +1321,36 @@ static int add_lco(struct msgb *msg, struct mgcp_msg *mgcp_msg)
 #undef MSGB_PRINTF_OR_RET
 }
 
+/* Helper function to obtain local IP address used in MGCP towards MGW,
+ * in string format, to fill the SDP "Origin" ("o=") field.
+ * return 0 on success, negative on error.
+ */
+static int get_mgcp_local_addr(const struct mgcp_client *mgcp, char *local_ip, size_t local_ip_len)
+{
+	int fd;
+
+	/* Try to get the socket local IP address if available: */
+	if (mgcp->iofd && ((fd = osmo_iofd_get_fd(mgcp->iofd)) >= 0)) {
+		if (osmo_sock_get_local_ip(fd, local_ip, local_ip_len) == 0)
+			return 0;
+		/* else: continue below */
+	}
+
+	/* If MGCP local address was explicitly specified in config, use it: */
+	if (mgcp->actual.local_addr) {
+		osmo_strlcpy(local_ip, mgcp->actual.local_addr, local_ip_len);
+		return 0;
+	}
+
+	/* Guess our local address based on system routing towards MGW: */
+	OSMO_ASSERT(local_ip_len >= INET6_ADDRSTRLEN);
+	if (osmo_sock_local_ip(local_ip, mgcp->actual.remote_addr) == 0)
+		return 0;
+
+	LOGPMGW(mgcp, LOGL_ERROR, "Could not determine local IP-Address!\n");
+	return -EINVAL;
+}
+
 /* Helper function for mgcp_msg_gen(): Add SDP information to MGCP message */
 static int add_sdp(struct msgb *msg, struct mgcp_msg *mgcp_msg, struct mgcp_client *mgcp)
 {
@@ -1316,6 +1360,7 @@ static int add_sdp(struct msgb *msg, struct mgcp_msg *mgcp_msg, struct mgcp_clie
 	const char *codec;
 	unsigned int pt;
 	uint16_t audio_port;
+	int rc;
 
 #define MSGB_PRINTF_OR_RET(FMT, ARGS...) do { \
 		if (msgb_printf(msg, FMT, ##ARGS) != 0) { \
@@ -1331,11 +1376,9 @@ static int add_sdp(struct msgb *msg, struct mgcp_msg *mgcp_msg, struct mgcp_clie
 	MSGB_PRINTF_OR_RET("v=0\r\n");
 
 	/* Determine local IP-Address */
-	if (osmo_sock_local_ip(local_ip, mgcp->actual.remote_addr) < 0) {
-		LOGPMGW(mgcp, LOGL_ERROR,
-			"Could not determine local IP-Address!\n");
-		return -EINVAL;
-	}
+	rc = get_mgcp_local_addr(mgcp, local_ip, sizeof(local_ip));
+	if (rc < 0)
+		return rc;
 	local_ip_family = osmo_ip_str_type(local_ip);
 	if (local_ip_family == AF_UNSPEC)
 		return -EINVAL;
@@ -1384,15 +1427,32 @@ static int add_sdp(struct msgb *msg, struct mgcp_msg *mgcp_msg, struct mgcp_clie
 	/* Add optional codec parameters (fmtp) */
 	if (mgcp_msg->param_present) {
 		for (i = 0; i < mgcp_msg->ptmap_len; i++) {
-			/* The following is only applicable for AMR */
-			if (mgcp_msg->ptmap[i].codec != CODEC_AMR_8000_1
-			    && mgcp_msg->ptmap[i].codec != CODEC_AMRWB_16000_1)
-				continue;
 			pt = mgcp_msg->ptmap[i].pt;
-			if (mgcp_msg->param.amr_octet_aligned_present && mgcp_msg->param.amr_octet_aligned)
-				MSGB_PRINTF_OR_RET("a=fmtp:%u octet-align=1\r\n", pt);
-			else if (mgcp_msg->param.amr_octet_aligned_present && !mgcp_msg->param.amr_octet_aligned)
-				MSGB_PRINTF_OR_RET("a=fmtp:%u octet-align=0\r\n", pt);
+			switch (mgcp_msg->ptmap[i].codec) {
+			case CODEC_AMR_8000_1:
+			case CODEC_AMRWB_16000_1:
+				if (!mgcp_msg->param.amr_octet_aligned_present)
+					break;
+				MSGB_PRINTF_OR_RET("a=fmtp:%u octet-align=%d\r\n",
+						   pt, (int)mgcp_msg->param.amr_octet_aligned);
+				break;
+			case CODEC_GSM_8000_1:
+			case CODEC_GSMEFR_8000_1:
+				if (!mgcp_msg->param.fr_efr_twts001_present)
+					break;
+				MSGB_PRINTF_OR_RET("a=fmtp:%u tw-ts-001=%d\r\n",
+						   pt, (int)mgcp_msg->param.fr_efr_twts001);
+				break;
+			case CODEC_GSMHR_8000_1:
+				if (!mgcp_msg->param.hr_twts002_present)
+					break;
+				MSGB_PRINTF_OR_RET("a=fmtp:%u tw-ts-002=%d\r\n",
+						   pt, (int)mgcp_msg->param.hr_twts002);
+				break;
+			default:
+				/* no parameters for the remaining codecs */
+				break;
+			}
 		}
 	}
 
